#!/usr/bin/perl
#	apt-cacher.pl - CGI to provide a local cache for debian packages and release files and .deb files
#
#  $Revision: 1.10 $
#  $Source: /xenu/nick/CVS-TREE/Src/Apt-cacher/apt-cacher.pl,v $
#  $Date: 2001/09/29 05:37:14 $
#
#	PATH_INFO=/www.domain.name/some/path/filename

=head1 NAME

 apt-cacher.pl - CGI to provide a cache for downloaded Debian packages

 Copyright (C) 2001, Nick Andrew <nick@zeta.org.au>
 Distributed under the terms of the GNU Public Licence (GPL).

=head1 SYNOPSIS

 copy apt-cacher.pl to your apache cgi-bin directory
 ./setup.pl /home/me/cache
 edit /etc/apt/sources.list
 apt-get update
 apt-get upgrade

=head1 DESCRIPTION

If you have two or more Debian GNU/Linux machines on a fast local
network and you wish to upgrade packages from the Internet, you
don't want to download every package several times.

apt-cacher.pl is a CGI which will keep a cache on disk of Debian Packages
and Releas files (including .debs) which have been received from Debian
distribution servers on the Internet. When an apt-get client issues
a request for a file to apt-cacher.pl, if the file is already on disk
it is served to the client immediately, otherwise it is fetched from
the Internet, saved on disk, and then served to the client. This means
that several Debian machines can be upgraded but each package need be
downloaded only once.

To use this CGI you need a web server which supports CGI and a local
directory with plenty of free space (100 - 200 Mbytes is usually
adequate).

=head1 INSTALLATION

Assuming your web server is called B<www.myserver.com:80>
and your cache directory is called B</home/me/cache>, then:

1. Copy apt-cacher.pl to your web server's cgi-bin directory

2. Make sure apt-cacher.pl is executable (chmod a+rx apt-cacher.pl)

3. Edit apt-cacher.pl and set $cache_dir to /home/me/cache

4. Make sure apt-cacher.pl is ok to run (perl -Mstrict -wc apt-cacher.pl)

5. Run B<./setup.pl /home/me/cache> to create necessary directories

6. Make sure your client machines can access http://www.myserver.com:80/cgi-bin/apt-cacher.pl

If the CGI is executed without arguments, it will return a text/plain
error message.

7. Edit your /etc/apt/sources.list files, as follows. Where a line says
something like:

deb http://http.us.debian.org/debian testing main contrib non-free

change this to:

deb http://www.myserver.com:80/cgi-bin/apt-cacher.pl/http.us.debian.org/debian testing main contrib non-free

8. Do "apt-get update" as root. This will prime the cache dorectory with the
Package or Package.gz and Release files from the servers you used to use
directly.

9. Do "ls -laR /home/me/cache" to verify that files have been received and
stored. The "/home/me/cache/tmp" directory should be empty after downloads
have completed.

10. Do "apt-get update; apt-get -u upgrade" to start upgrading each machine.

=head1 CACHE DIRECTORY CONTENTS

apt-cacher.pl considers all .deb files with exactly the same filename
should be the same package (for example vim-rt_5.8.007-4_all.deb) no
matter where they are downloaded from, so these files are stored in
the cache directory using just the filename.

Packages and Release files (including Packages.gz) are potentially
different for every server and directory, so these files are stored
in the cache directory with the full hostname and path to the file,
with all slashes B</> changed to underscores B<_> (in the same
manner as apt-get names the files in B</var/lib/apt/lists>).

=head1 BUGS and FEATURES

1. Only HTTP is supported at present (i.e. apt-cacher.pl cannot access an
FTP URL)

2. apt-cacher.pl probably only works with the Apache webserver, because
it relies on the webserver supplying the PATH_INFO environment variable

3. apt-cacher.pl uses B<wget> to retrieve files, so wget must be
installed.

4. (this bug has been squashed)

5. (this bug has been squashed)

6. (this bug has been squashed)

7. The cache directory is never cleaned out by this tool, and it does
not issue an HTTP GET-IF-MODIFIED-SINCE request for Packages{,.gz}
and Release files, which may be frequently modified, so the user should
delete these files from the cache occasionally:

  find $cache_dir/tmp -type f -ctime +1 -print -exec rm {} \;

  find $cache_dir -type f \( -name Package\* -o -name Release\* \) -ctime +7 -print -exec rm {} \;

8. apt-get can resume a partial failed transfer, however apt-cacher.pl
cannot.

9. When wget is used with the '-c' option (to continue a partial transfer)
and the '-s' option (to write HTTP response headers to the file) it writes
the second response headers into the middle of the partial file. So we
can't use continue mode, so we have to truncate any existing file before
trying again.

=head1 ENVIRONMENT VARIABLES

B<PATH_INFO> is used to find the full URL for the requested file

B<PATH> is used to search for the wget executable

=head1 UPDATES

Please email bug fixes and enhancements to B<nick@zeta.org.au> using
B<apt-cacher> in the subject line. All reasonable changes will be
included in the next release.

The homepage for apt-cacher is
B<http://www.zeta.org.au/~nick/projects/apt-cacher/>.
Please check it occasionally for updates.

=cut
# ----------------------------------------------------------------------------

my $path = $ENV{PATH_INFO};
my $cache_dir = "/xenu/home/nick/tmp/apt-cache";
my $private_dir = "$cache_dir/private";
my $do_lock = 0;
my $debug = 0;

# use IO::Handle;
use Fcntl ':flock';
use IO::Handle;
use POSIX;

# Output data as soon as we print it
$| = 1;

# ----------------------------------------------------------------------------
# We have a problem for large packages. The problem is that it appears
# the apache webserver gets tired of waiting for us to return a document
# so it sends a SIGTERM and then soon afterward, a SIGKILL. This is
# apparently a configuration directive in mod_perl ?? Anyway, let's fork
# a child so apache's violence doesn't kill us...

if (fork() > 0) {
	# parent
	$SIG{'TERM'} = sub { print STDERR "--- apt-cacher.pl: Parent received SIGTERM, exiting\n"; sleep(4); exit(8); };
	wait();
	exit($?);
}

# ----------------------------------------------------------------------------
# Data also used by child processes

my $unique_filename;
my $child_pid;
my $child_rc;

# ----------------------------------------------------------------------------

# $SIG{'PIPE'} = sub { open(E, ">>$cache_dir/errs"); print E "$$ received SIGPIPE\n"; close(E); };
# $SIG{'PIPE'} = 'IGNORE';
my $sigpipe_received = 0;

$SIG{'PIPE'} = sub {
	print STDERR "--- apt-cacher.pl: received SIGPIPE\n";
	$sigpipe_received = 1;
};

sub term_handler {
	print STDERR "--- apt-cacher.pl: received SIGTERM, terminating\n";

	# Kill the wget process if running and unlink its output file

	kill('TERM', $child_pid) if ($child_pid);
	unlink($unique_filename) if ($unique_filename);
	exit(8);
};

$SIG{'TERM'} = \&term_handler;

$SIG{'QUIT'} = sub { print STDERR "--- apt-cacher.pl: received SIGQUIT\n"; };
$SIG{'INT'} = sub { print STDERR "--- apt-cacher.pl: received SIGINT\n"; };

$SIG{'CHLD'} = sub {
	print STDERR "--- apt-cacher.pl: received SIGCHLD\n";
	wait();
	$child_rc = $?;
	undef $child_pid;
};

print STDERR "\n--- apt-cacher.pl: called with $path\n";

die "apt-cacher.pl: No cache_dir directory!\n" if (!-d $cache_dir);
die "apt-cacher.pl: No cache_dir/tmp directory!\n" if (!-d "$cache_dir/tmp");
die "apt-cacher.pl: No cache_dir/private directory!\n" if (!-d "$cache_dir/private");

$debug = 1 if (-f "$cache_dir/debug");

if ($do_lock) {
	open(LOCK, ">$cache_dir/lock") or die "apt-cacher.pl: Unable to open $cache_dir/lock for write: $!\n";
	if (!flock(LOCK, LOCK_EX)) {
		die "Unable to achieve lock on $cache_dir/lock: $!\n";
	}

	print STDERR "--- apt-cacher.pl: Lock achieved\n";
}

# keep LOCK open so that at most one apt-cacher.pl can be running at any time

# Now parse the path
if ($path !~ m(^/.+/.+)) {
	usage_error();
	exit(4);
}

my($host,$uri) = ($path =~ m(^/([^/]+)(/.+)));

if ($host eq '' || $uri eq '') {
	usage_error();
	exit(4);
}

my ($filename) = ($uri =~ /\/([^\/]+)$/);
my $new_filename;

if ($filename =~ /\.deb$/) {
	# Place the file in the cache with just its basename
	$new_filename = $filename;
} else {
	# Make a long filename so we can cache these files
	$new_filename = "$host$uri";
	$new_filename =~ s/\//_/g;
}

my $cached_file = "$cache_dir/$new_filename";

print STDERR "--- Looking for $cached_file\n";

my $is_open = 0;	# Is the file currently open by us?
my $is_incomplete = 0;	# Is the file contents complete?

if (!-f $cached_file) {
	# File does not exist, so try to create it
	# KLUDGE ... probably a race condition here
	unlink("$private_dir/$new_filename.complete");
	print STDERR "--- File does not exist, create it\n";
	if (sysopen(CF, $cached_file, O_RDWR|O_CREAT|O_EXCL, 0644)) {
		$is_open = 1;
	}
	# If open fails, maybe we came 2nd in a race
	# ... KLUDGE ... continue here
}

if (!-f $cached_file) {
	barf("Tried to create $cached_file, but failed");
}


# Ok, the file exists. Open it if we didn't already.
if (!$is_open) {
	print STDERR "--- Open $cached_file\n";

	if (!sysopen(CF, $cached_file, O_RDWR)) {
		barf("Unable to open incomplete $cached_file: $!");
	}
	$is_open = 1;
}

# Is it incomplete?
if (!-f "$private_dir/$new_filename.complete") {
	$is_incomplete = 1;
	print STDERR "--- File is not complete\n";

	if (!flock(CF, LOCK_EX|LOCK_NB)) {
		# flock failed, assume fetcher is running already
		print STDERR "--- Unable to lock, fetcher must be running\n";
		# KLUDGE ... race condition, wait for fetcher to start up
		sleep(3);
	} else {
		# file locked, nobody's touching it ...
		# Have to truncate it, because we can't rely on "resume"
		truncate(CF, 0);

		try_pickup($host, $uri, $cached_file, $new_filename);
	}
}

# At this point the file is open, and it's either complete or somebody
# is fetching its contents


print STDERR "--- Starting to return $cached_file\n";

my $first_line = 1;
my($buf,$n,$bufsize);
my $abort_timer = 300;
my $nodata_count = 0;

while (1) {
	if ($sigpipe_received) {
		print STDERR "--- Exit (SIGPIPE)\n";
		exit(4);
	}

	$n = read(CF, $buf, 65536);
	barf("Oops, read failed!") if (!defined $n);

	print STDERR "Read $n bytes\n" if ($debug);

	if ($n < 0) {
		print STDERR "--- Exit (read fail)\n";
		exit(4);
	}

	# if ($n == 0), there's no more data to read yet
	if ($n == 0) {
		if (!$is_incomplete || ($nodata_count > 0 && -f "$private_dir/$new_filename.complete")) {
			# Looks like file is complete!
			# Finish up
			print STDERR "--- Exit (file completed)\n";
			last;
		}

		$nodata_count += 2;
		if ($nodata_count >= $abort_timer) {
			print STDERR "--- Abort (timeout)\n";
			exit(4);
		}
		sleep(2);
		next;
	}

	$nodata_count = 0;

	# Hey, there's data! Is it the first line?
	if ($first_line) {
		my $i = index($buf, "\n");
		if ($i >= 0) {
			# Throw away first line only
			$buf = substr($buf, $i + 1);
			# Output remainder, if any
			if (length $buf) {
				print $buf;
				print STDERR "Wrote initial ", length($buf), " bytes\n" if ($debug);
			}
			$first_line = 0;
		} else {
			# End of first line not found, throw it all away
		}
	} else {
		print $buf;
		print STDERR "Wrote ", length($buf), " bytes\n" if ($debug);
	}
}

# We're done!
exit(0);

sub barf {
	my $errs = shift;

	die "--- apt-cacher.pl: Fatal: $errs\n";
}

sub usage_error {
	print STDERR "--- apt-cacher.pl: Usage error\n";

	print <<EOF;
Content-Type: text/plain
Expires: 0

Usage: GET .../apt-cacher.pl/www.domain.name/some/path/to/filename
EOF

}

sub try_pickup {
	my $host = shift;
	my $uri = shift;
	my $cached_file = shift;
	my $new_filename = shift;

	# Otherwise, try to pick it up in the background ...

	# Set this in the parent so the parent knows what file to delete
	# if it receives a SIGTERM
	$unique_filename = "$cache_dir/$new_filename";

	my $pid = fork();
	if ($pid < 0) {
		barf("fork() failed");
	}

	if ($pid > 0) {
		# parent
		return;
	}

	# child

	my $url = "http://$host$uri";

	print STDERR "--- Fetcher: Try to pick up $url\n";
	$child_pid = fork();
	barf("Fetcher: fork failed") if (!defined $child_pid);

	if ($child_pid == 0) {
		# This is the child

		# Search the PATH environment for the wget executable
		foreach my $dir (split(/:/, $ENV{PATH})) {
			my $fn = "$dir/wget";
			if (-f $fn && -x _) {
				# { } surround exec to prevent perl warning
				{
					# had to remove the -c option from wget because resuming is incompatible with the -s option
					exec($fn, '-s', '-O', $unique_filename, $url);
				};
			}
		}

		# dang, no wget in path
		barf("Unable to exec wget");
	}

	# Otherwise, we must be the parent

	# Child processes are reaped by signal handler
	while ($child_pid > 0) {
		pause();
	}

	my $rc = $child_rc;

	if ($rc != 0) {
		unlink($unique_filename);
	}

	print STDERR "--- Pick up $url as $unique_filename, return code $rc\n";
	if ($rc != 0) {
		# Output nothing and exit - this will be a 500 server error?
		exit(0);
	}

	# Now create the file to show the pickup is complete
	open(MF, ">$cache_dir/private/$new_filename.complete");
	close(MF);

	print STDERR "--- Fetcher exiting\n";

	exit(0);
}
